# -*- coding: utf-8 -*-
# file: evolution.py
# time: 2021/7/30
# author: yangheng <yangheng@m.scnu.edu.cn>
# github: https://github.com/yangheng95
# Copyright (C) 2021. All Rights Reserved.
import random

import numpy as np

from EvoInjection.deep_learning.classifier.fitness_assigner import FitnessAssigner
from EvoInjection.deep_learning.translator.translator import InjectionTranslator
from EvoInjection.evolution.entity.individual import Individual
from EvoInjection.evolution.entity.population import Population
from EvoInjection.utils.bypass import is_bypass

from EvoInjection.global_config import MODSECURITY_SERVER_IP, OPENRESTY_SERVER_IP


class SingleTaskEvolution:
    def __init__(self, args, logger):
        self.task = args.tasks[0]
        self.pop = Population(self.task, args.pop_size)
        self.bypass_injection_by_task = {task: set() for task in args.tasks}
        self.count_by_task = {task: 0 for task in args.tasks}

        self.fitness_assigner = FitnessAssigner(self.task, logger.classifier)
        self.fitness_assigner.assign_fitness(self.pop)

        self.logger = logger
        self.waf_address = MODSECURITY_SERVER_IP if args.waf == 'mod_security' else OPENRESTY_SERVER_IP

        self.visited = set()

    def evolve(self, gen_id, rnd_select):
        avg_fitness = self.pop.get_average_fitness()
        _pop = self.pop[:]
        for i in range(len(self.pop)):
            idv = self.pop[i].mutate()
            while idv.injection in self.visited:
                # continue
                idv = Individual(idv.task)

            self.visited.add(idv.injection)

            self.fitness_assigner.assign_fitness(idv)

            if is_bypass(idv, self.waf_address) and idv.injection not in self.bypass_injection_by_task[self.task]:
                self.count_by_task[self.task] += 1
                self.bypass_injection_by_task[self.task].add(idv.injection)
            if not rnd_select:
                if idv.fitness > avg_fitness:
                    _pop.append(idv)
                else:
                    _pop.append(Individual(idv.task))
        if not rnd_select:
            self.fitness_assigner.assign_fitness(_pop)
            _pop = sorted(_pop, key=lambda x: x.fitness, reverse=True)

        self.pop.individuals = _pop[:len(self.pop)]
        self.logger.log_count(self.count_by_task, gen_id)


class MultiTaskEvolution:
    def __init__(self, args, logger):
        self.tasks = args.tasks
        self.pops = {t: Population(t, args.pop_size) for t in args.tasks}
        self.bypass_injection_by_task = {task: set() for task in args.tasks}
        self.count_by_task = {task: 0 for task in args.tasks}

        self.translators = {}
        for t1 in self.tasks:
            for t2 in self.tasks:
                if t1 != t2:
                    self.translators['{}2{}'.format(t1, t2)] = InjectionTranslator(t1, t2)

        self.fitness_assigners = {t: FitnessAssigner(t, logger.classifier) for t in self.tasks}

        for task in self.tasks:
            self.fitness_assigners[task].assign_fitness(self.pops[task])

        self.logger = logger
        self.waf_address = MODSECURITY_SERVER_IP if args.waf == 'mod_security' else OPENRESTY_SERVER_IP

        self.visited = set()

    def evolve(self, gen_id, rnd_select):

        for pop_name in self.pops:
            avg_fitness = self.pops[pop_name].get_average_fitness()
            _pop = self.pops[pop_name][:]
            for i in range(len(self.pops[pop_name])):
                if random.random() < max(np.exp(- gen_id), 0.01):
                    idv = self.translate(pop_name)
                else:
                    idv = self.pops[pop_name][i].mutate()
                self.fitness_assigners[pop_name].assign_fitness(idv)

                while idv.injection in self.visited:
                    idv = Individual(idv.task)
                self.visited.add(idv.injection)

                if is_bypass(idv, self.waf_address) and idv.injection not in self.bypass_injection_by_task[pop_name]:
                    self.count_by_task[pop_name] += 1
                    self.bypass_injection_by_task[pop_name].add(idv.injection)

                if not rnd_select:
                    if idv.fitness > avg_fitness:
                        _pop.append(idv)
                    else:
                        _pop.append(Individual(idv.task))

            if not rnd_select:
                self.fitness_assigners[pop_name].assign_fitness(_pop)
                _pop = sorted(_pop, key=lambda x: x.fitness, reverse=True)
            self.pops[pop_name].individuals = _pop[:len(self.pops[pop_name])]

        self.logger.log_count(self.count_by_task, gen_id)

    def translate(self, pop_name):
        _other_tasks_ = self.tasks[:]
        _other_tasks_.remove(pop_name)
        select_task_name = random.choice(_other_tasks_)
        idx = random.randint(0, 99)
        # translate
        translator_name = '{}2{}'.format(select_task_name, pop_name)
        idv = Individual(pop_name)
        idv.injection = self.translators[translator_name].translate(self.pops[select_task_name][idx].injection)
        return idv
